use std::{env::consts::DLL_EXTENSION, fs::create_dir_all, path::Path, process::ExitCode};

use clap::Args;
use git2::Repository;
use tempfile::tempdir;

use crate::console::{GOOD, NOTE, error, fail, status, tip, update, warning};
use crate::settings::{InstallSettings, Settings};

mod cc;
pub(crate) mod languages;

use languages::{InstallInfo, Revision, is_installed};

#[derive(Debug, Args)]
pub struct Install {
    /// Reinstall already present parsers
    #[arg(short, long)]
    force: bool,

    /// List installable parsers
    #[arg(short, long)]
    list: bool,

    languages: Vec<String>,
}

impl Install {
    pub fn run(self, mut settings: Settings) -> anyhow::Result<ExitCode> {
        let language_defs = settings.language_defs()?;

        if self.list {
            let mut available = language_defs.into_keys().collect::<Vec<_>>();
            available.sort();
            available.dedup();

            for language in available {
                println!("{language}");
            }

            return Ok(ExitCode::SUCCESS);
        }

        let src_dir = tempdir()?;

        let parser_dir = settings.parser_dir();
        let InstallSettings { compiler } = settings.install;

        create_dir_all(&parser_dir)?;

        // have to use a canonical path bc compiler runs elsewhere
        let parser_dir = parser_dir.canonicalize()?;

        let Self {
            mut languages,
            force,
            ..
        } = self;

        languages.sort();
        languages.dedup();

        let mut errors = 0;
        let mut skips = 0;

        for language in &languages {
            if !force && is_installed(language, &parser_dir) {
                status!("Skipping", &NOTE, "{language} (already installed)");
                skips += 1;
            } else if let Some(info) = language_defs.get(language) {
                if let Err(Error::Fatal(_)) = install_with_status(
                    language,
                    info,
                    &parser_dir,
                    src_dir.as_ref(),
                    compiler.as_deref(),
                ) {
                    return Ok(ExitCode::FAILURE);
                } else {
                    errors += 1;
                }
            } else {
                error!("unknown language {language}");
                errors += 1;
            }
        }

        if skips > 0 {
            tip!("try '{NOTE}--force{NOTE:#}' to reinstall parsers");
        }

        Ok(if errors > 0 {
            ExitCode::FAILURE
        } else {
            ExitCode::SUCCESS
        })
    }
}

fn install_with_status(
    language: &str,
    install_info: &InstallInfo,
    parser_dir: &Path,
    src_dir: &Path,
    compiler: Option<&str>,
) -> Result<Option<Revision>, Error> {
    status!("Installing", &GOOD, "{language}");

    let res = install(language, install_info, parser_dir, src_dir, compiler);

    match &res {
        Err(e) => fail!("{e}"),
        Ok(Some(revision)) => update!("Installed", &GOOD, "{language} {revision:.7}"),
        Ok(None) => {
            update!("Installed", &GOOD, "{language}");
            warning!("couldn't get version info for {language}")
        }
    }

    res
}

#[derive(Debug, thiserror::Error)]
enum Error {
    #[error("couldn't clone `{0}`")]
    Clone(String),

    #[error(transparent)]
    Git(#[from] git2::Error),

    #[error(transparent)]
    Compile(cc::Error),

    #[error(transparent)]
    Fatal(cc::Error),
}

impl From<cc::Error> for Error {
    fn from(e: cc::Error) -> Self {
        use cc::Error as E;

        match e {
            E::NoCompiler | E::BadCompiler { .. } | E::EmptyCompiler => Self::Fatal(e),
            _ => Self::Compile(e),
        }
    }
}

fn install(
    language: &str,
    InstallInfo { url, files }: &InstallInfo,
    parser_dir: &Path,
    src_dir: &Path,
    compiler: Option<&str>,
) -> Result<Option<Revision>, Error> {
    let repo_path = src_dir.join(language);
    let repo = Repository::clone(url, &repo_path).map_err(|_| Error::Clone(url.to_owned()))?;

    // it's not fatal if we just can't get the revision info, tho it'd definitely be weird
    let version = repo
        .head()
        .and_then(|head| head.peel_to_commit().map(|c| c.id()))
        .ok();

    let output_path = parser_dir.join(language).with_extension(DLL_EXTENSION);

    let mut build = cc::Build::new();
    build.compiler(compiler).dir(&repo_path).input_files(files);
    build.compile(&output_path)?;

    Ok(version)
}
