use std::{
    path::{Path, PathBuf},
    process::ExitCode,
};

use clap::{Parser, Subcommand};
use clap_cargo::style::CLAP_STYLING;

mod console;
mod install;
mod settings;

use console::{Color, error};
use install::Install;
use settings::{default_config, get_settings};

#[derive(Debug, Parser)]
#[command(version, about, long_about = None)]
#[command(styles = CLAP_STYLING)]
struct Cli {
    /// Path to config file
    #[arg(short, long, global = true, value_name = "PATH")]
    config: Option<PathBuf>,

    /// When to use terminal colors
    #[arg(long, value_enum, global = true, default_value_t = Color::Auto, value_name="WHEN")]
    color: Color,

    #[command(subcommand)]
    command: Command,
}

#[derive(Debug, Subcommand)]
enum Command {
    /// Install parsers
    Install(Install),
}

impl Command {
    fn run(self, config_path: &Path) -> anyhow::Result<ExitCode> {
        let settings = get_settings(config_path)?;

        match self {
            Self::Install(args) => args.run(settings),
        }
    }
}

fn main() -> ExitCode {
    let Cli {
        config,
        color,
        command,
    } = Cli::parse();

    console::init(color);

    command
        .run(&config.unwrap_or_else(default_config))
        .unwrap_or_else(|e| {
            error!("{e}");
            ExitCode::FAILURE
        })
}
